#include "stdafx.h"
#include "process.h"
#include "math.h"
#include "conio.h"

#include <vector>
using std::vector;

// todo - OnBtnRecord calc from miniscroll settings and acq rate 
// todo - OnBtnRecord set buffer size in samples - enough for a few seconds.
// todo - OnBtnRecord if miniscroll sets SampPerPix too high for queue, prevent 
// todo - prevent creation of session if no hardware configs exist
// todo - allow task to be created ready to go, then start it when user clicks.

/* The following ID's are used for the manually created controls ( live modify ): 
	prot [Title]	ID_GENCTL + (iCh+1)*ID_GENCTLX + 0
	prot [Iter Lbl]	ID_GENCTL + (iCh+1)*ID_GENCTLX + 1
	prot [Iter edit]ID_GENCTL + (iCh+1)*ID_GENCTLX + 2
	prot [Update]	ID_GENCTL + (iCh+1)*ID_GENCTLX + 3
	prot [Play]		ID_GENCTL + (iCh+1)*ID_GENCTLX + 4
	prot [Loop]		ID_GENCTL + (iCh+1)*ID_GENCTLX + 5
	prot [Stop]		ID_GENCTL + (iCh+1)*ID_GENCTLX + 6
	pid(p) label	ID_GENCTL + (iCh+1)*ID_GENCTLX + 7
	pid(p) value	ID_GENCTL + (iCh+1)*ID_GENCTLX + 8
	pid(i) label	ID_GENCTL + (iCh+1)*ID_GENCTLX + 9
	pid(i) value	ID_GENCTL + (iCh+1)*ID_GENCTLX + 10
	pid(d) label	ID_GENCTL + (iCh+1)*ID_GENCTLX + 11
	pid(d) value	ID_GENCTL + (iCh+1)*ID_GENCTLX + 12

	seg label		ID_GENCTL + (iCh+1)*ID_GENCTLX + (1+iPnlSeg)*ID_GENCTLPNLX + 1
	seg dur			ID_GENCTL + (iCh+1)*ID_GENCTLX + (1+iPnlSeg)*ID_GENCTLPNLX + 2
	u label			ID_GENCTL + (iCh+1)*ID_GENCTLX + (1+iPnlSeg)*ID_GENCTLPNLX + 3
	u value			ID_GENCTL + (iCh+1)*ID_GENCTLX + (1+iPnlSeg)*ID_GENCTLPNLX + 4
	v label			ID_GENCTL + (iCh+1)*ID_GENCTLX + (1+iPnlSeg)*ID_GENCTLPNLX + 5
	v value			ID_GENCTL + (iCh+1)*ID_GENCTLX + (1+iPnlSeg)*ID_GENCTLPNLX + 6
	u range			ID_GENCTL + (iCh+1)*ID_GENCTLX + (1+iPnlSeg)*ID_GENCTLPNLX + 7
	v range			ID_GENCTL + (iCh+1)*ID_GENCTLX + (1+iPnlSeg)*ID_GENCTLPNLX + 8
	
	Digital Outputs use iCh=16 
*/				

const int ID_GENCTL=5000;		// Base ID for runtime created controls in generation panels 
const int ID_GENCTLX=1024;		// Range of ID's per gen
const int ID_GENCTLPNLX=16;		// Range of ID's per gen panel

const int ID_DIGGENCTL=ID_GENCTL+16*ID_GENCTLX;	//base ID for Digital Generation Controls 
const int ID_DIGACQCTL=ID_GENCTL+17*ID_GENCTLX;	//base ID for Digital Acquisition Controls 
								//  (leaves room for 16 analog gen panels)

const int UPDATE_WIDTH=20;		// Width of [1] and [*] buttons on gen panel 
const int EDIT_WIDTH=75;		// Width of value edit on gen panel +10 for spacing

const int ID_VBAR = 2001;		// sets a range 2001..2016
const int ID_TITLE = 2051;		// sets a range 2001..2016
const int ID_UNITS = 2101;		// sets a range 2001..2016

const int MINSPEED = 1;			// range of values for slider that controls time/div scroll speed.
const int MAXSPEED = 500;		//	a.k.a. samples per pixel

const int ACQ_QUE_SIZE = 10000;		//todo - user modifiable ? 

const int DIGOUTCTL_BUFSIZE = 1024;

const UINT WM_FINISHED_RESIZING_VWSESS = WM_USER + 8251;

//-----------------------------------------------------------------------
class CVwSession : public CDialog { 
	public :
		CSession m_oSession;
		RECT m_rctDisp;				// coordinates of display area 
		RECT m_rctFocus;
		CString m_cLongName;		// file name - blank on New sessions
		int m_iDrawMode;			// 3=every pt. 2=avg(pts) 1=1 point 
		int m_iScanMode;			// 1=scrolling rt to lft. 2=tracing Lft to rt.
		int m_iAcqMode;				// current acquisition mode = MODE_IDLE , MODE_SYNCH, ... 
		int m_iGenMode;				// current generation mode MODE_IDLE , MODE_SYNCH, ... 

		bool m_bSetup;				// session has been loaded or created 
		int m_iQuePos;				// position in session que last copied
		double m_dQuePos;			// For the next bit, starting offset in Session Que to read 
									// Note use of Que, not Queue to distinguish from Session:Queue
		int m_iPage, m_iPages;		// Display page for multipage views
								
		int m_iElapsed;				// samples since start of rec.
		int m_iTimeBar;				// Elapsed at last time par set.
		int m_iTimeSep;				// samples between time bars
		bool m_bTimeSep;			// toggle whether time bars appear
		bool m_bResizing;           // whether to ignore text-changed events for u,v
				
		double m_dTmMin;			// time axis Min
		double m_dTmMax;			// time axis Max
		double m_dTm1st;			// time axis 1st line
		double m_dTmSep;			// time axis sep btw lines.

		int m_aiCapture[6];			// The panel, x, y when mouse capture for sizing panels begins
									// or, window width, height between WM_SIZE and WM_EXITSIZEMOVE

		char * m_aDigoutControl;	// Control structure for communication with digital polling thread
		
		//--- windows controls 
		CDlgCheck m_chkDigout[CONFIG_DOROWS];	// checkbox : Manually activated digital output 

		CDlgControl m_picDisp;		// picture : Scope display area
		CDlgControl m_sliderTimeDiv;// slider : display scroll rate 
		CDlgButton m_btnPage;		// Page selection
		
		CDlgButton m_btnPlay;		// button : Acq start play mode
		CDlgButton m_btnRecord;		// button : acq start record mode
		CDlgButton m_btnStop;		// button : stop acquisition or synchronized acq/gen

		CDlgButton m_btnSynch;		// button : Start synchronized acquisition / generation

		CDlgButton m_btnGenAll;		// button : Start all generations
		CDlgButton m_btnGenAllStop;	// button : Stop generations started by GenAll
		
		CDlgButton m_btnDrawMode[3];// button : Fast/medium/slow drawing detail level

		CDlgCheck m_btnScanMode[2];	// button : Oscilloscope or chart recorder style scrolling

		CDlgCheck m_btnTimeSep;		// button : Hide/show vertical time separation bars
		
		CDlgButton m_btnNote;		// button : Add a data note at current recording position.
		CDlgButton m_btnConfig;		// button : Change session configuration - disabled 
		CDlgButton m_btnSave;		// button : Save Session settings.
		CDlgButton m_btnClose;		// button : Close session - without save
		CDlgButton m_btnSaveClose;	// button :	Save session settings and close session
		CDlgButton m_btnQub;		// button : Close data file and open with QuB program
				
		CDlgControl m_lblElapsed;	// label : Show elapsed samples recorded count

		bool m_bMaximized;

		CVwSession() { 
			m_bSetup=false;
			m_bTimeSep=true;
			m_iID = IDD_SESSIONVIEW;
			m_dTmMin=0;
			m_dTmMax=5000;
			m_dTm1st=0;
			m_dTmSep=1000;
			m_pCapture=NULL;
			m_cLongName.Set("");
			m_iDrawMode=3;
			SetScanMode( 1 );
			SetElapsed(0);
			m_iTimeSep=1000;	//todo - set appropriately 
			m_iTimeBar=0;
			m_aiCapture[0]=NOCAPTURE;
			m_aDigoutControl = NULL;
			m_iPage=0;
			m_iPages=1;
			m_bMaximized = false;
			SetRect( &m_rctFocus, -1, -1, -1, -1 );
			}
			
		~CVwSession() { 
			}

		//----------			
		void InitializeControls( ) { 
			//-- digital output controls
			char buf[100];
			for( int i=0; i<CONFIG_DOROWS; ++i ) { 
				sprintf( buf, "Toggle digital output %d", i);
				m_chkDigout[i].SetHandle( GetDlgItem( m_hDlg, IDC_DIGOUT1+i), buf );	
				}

			//-- Global polling buttons
			m_picDisp.SetHandle( GetDlgItem( m_hDlg, IDC_DISP ) );
			m_picDisp.GetControlBounds( &m_rctDisp );
			
			m_sliderTimeDiv.SetHandle( GetDlgItem( m_hDlg, IDC_TIMEDIV ), "Samples per pixel" );
			m_btnPlay.SetHandle( GetDlgItem( m_hDlg, IDC_BTNPLAY ), "Start display w/o record" );
			m_btnRecord.SetHandle( GetDlgItem( m_hDlg, IDC_BTNRECORD ), "Start display w/ record" );
			m_btnStop.SetHandle( GetDlgItem( m_hDlg, IDC_BTNSTOP ), "Stop acquisition" );
			m_btnSynch.SetHandle( GetDlgItem( m_hDlg, IDC_BTNSYNCH ), "Start record and generate" );
			m_btnGenAll.SetHandle( GetDlgItem( m_hDlg, IDC_BTNGENALL ), "Start all generations" );
			m_btnGenAllStop.SetHandle( GetDlgItem( m_hDlg, IDC_BTNGENALLSTOP ), "Stop all generations" );

			//-- Chart appearance 
			m_btnDrawMode[0].SetHandle( GetDlgItem( m_hDlg, IDC_BTNDRAWMODE1 ), "Plot 1 sample per column" );
			m_btnDrawMode[1].SetHandle( GetDlgItem( m_hDlg, IDC_BTNDRAWMODE2 ), "Plot average sample per column" );
			m_btnDrawMode[2].SetHandle( GetDlgItem( m_hDlg, IDC_BTNDRAWMODE3 ), "Plot every sample" );
			m_btnScanMode[0].SetHandle( GetDlgItem( m_hDlg, IDC_BTNSCAN_MODE1 ), "Scrolling scope display" );
			m_btnScanMode[1].SetHandle( GetDlgItem( m_hDlg, IDC_BTNSCAN_MODE2 ), "Overwrite scope display" );
			m_btnTimeSep.SetHandle( GetDlgItem( m_hDlg, IDC_BTNTIMESEP ), "Toggle drawing time divisions" );
			m_btnTimeSep.SetCheck( m_bTimeSep );
			m_btnPage.SetHandle( GetDlgItem( m_hDlg, IDC_BTNPAGE ), "Page selection" );

			SendMessage( GetDlgItem(m_hDlg,IDC_BTNDRAWMODE1), BM_SETIMAGE, IMAGE_BITMAP, (LPARAM) g_hbDraw[0] );

			//-- save / close
			m_btnNote.SetHandle( GetDlgItem( m_hDlg, IDC_BTNNOTE ), "Add a note at the current sample" );
			m_btnConfig.SetHandle( GetDlgItem( m_hDlg, IDC_BTNCONFIG ), "Set rates and file names" );
			m_btnSave.SetHandle( GetDlgItem( m_hDlg, IDC_BTNSAVE ), "Save session settings" );
			m_btnSaveClose.SetHandle( GetDlgItem( m_hDlg, IDC_BTNSAVEANDCLOSE ), "Save session settings and close session" );
			m_btnClose.SetHandle( GetDlgItem( m_hDlg, IDC_BTNCLOSE ), "Close session w/o saving" );
			m_btnQub.SetHandle( GetDlgItem( m_hDlg, IDC_BTNQUB ), "close session and open data file with Qub" );
			m_lblElapsed.SetHandle( GetDlgItem( m_hDlg, IDC_LBLTIMEELAPSED ) ); 

			//-- set button bitmaps for buttons not added as controls
			SendMessage( GetDlgItem(m_hDlg,IDC_BTNDRAWMODE1), BM_SETIMAGE, IMAGE_BITMAP, (LPARAM) g_hbDraw[0] );
			SendMessage( GetDlgItem(m_hDlg,IDC_BTNDRAWMODE2), BM_SETIMAGE, IMAGE_BITMAP, (LPARAM) g_hbDraw[1] );
			SendMessage( GetDlgItem(m_hDlg,IDC_BTNDRAWMODE3), BM_SETIMAGE, IMAGE_BITMAP, (LPARAM) g_hbDraw[2] );

			SendMessage( GetDlgItem(m_hDlg,IDC_BTNSCAN_MODE1), BM_SETIMAGE, IMAGE_BITMAP, (LPARAM) g_hbScan[0] );
			SendMessage( GetDlgItem(m_hDlg,IDC_BTNSCAN_MODE2), BM_SETIMAGE, IMAGE_BITMAP, (LPARAM) g_hbScan[1] );
			
			//-- set modes to IDLE 
			SetModes( MODE_IDLE, MODE_IDLE );
			}
			
		//----- 
		void SetModes( int iAcqMode, int iGenMode ) { 
			//-- Set selected modes
			if( iAcqMode != MODE_NULL ) 
				m_iAcqMode=iAcqMode;
			if( iGenMode != MODE_NULL ) 
				m_iGenMode=iGenMode;

			//-- Page selection 
			if( m_iPages > 1 ) 
				m_btnPage.SetButtonBitmap( g_hbPage );
			else 
				m_btnPage.SetVisible( false );
			
			//-- set button display bitmaps based on whether each option is enabled in current mode
			m_btnPlay.SetButtonBitmap( g_hbPlay[ m_iAcqMode==MODE_IDLE || m_iAcqMode==MODE_ACQRECORD ? 0 : 1 ] );
			m_btnRecord.SetButtonBitmap( g_hbRecord[ m_iAcqMode==MODE_IDLE || m_iAcqMode==MODE_ACQPLAY ?  0 : 1 ] );
			m_btnStop.SetButtonBitmap( g_hbStop[ m_iAcqMode!=MODE_IDLE ? 0 : 1 ] );
			m_btnSynch.SetButtonBitmap( g_hbSynch[ m_iAcqMode==MODE_IDLE && m_iGenMode==MODE_IDLE ? 0 : 1 ] );
			m_btnGenAll.SetButtonBitmap( g_hbPlay[ m_iAcqMode!=MODE_SYNCH && m_iGenMode==MODE_IDLE ? 2 : 1 ] );
			m_btnGenAllStop.SetButtonBitmap( g_hbStop[ m_iAcqMode!=MODE_SYNCH && m_iGenMode==MODE_GENALL ? 2 : 1 ] );
			m_btnTimeSep.SetButtonBitmap( g_hbTimeSep[ m_bTimeSep ? 0 : 1 ] );
			
			//-- set button display bitmaps on generation panels 
			SetGenButtonBitmaps( m_iAcqMode, m_iGenMode );
				
			//-- Allow / prevent configuration changes
			//TODO - Add a resetup option to SetupSession before enabling this.
			//m_btnConfig.SetEnabled( m_iAcqMode==MODE_IDLE && m_iGenMode==MODE_IDLE );
			m_btnConfig.SetEnabled(false);
			
			//-- set session save / close buttons bitmaps
			m_btnSave.SetEnabled( m_iAcqMode==MODE_IDLE && m_iGenMode==MODE_IDLE );
			m_btnSaveClose.SetEnabled( m_iAcqMode==MODE_IDLE && m_iGenMode==MODE_IDLE );
			m_btnClose.SetEnabled( m_iAcqMode==MODE_IDLE && m_iGenMode==MODE_IDLE );
			m_btnQub.SetEnabled( m_iAcqMode==MODE_IDLE && m_iGenMode==MODE_IDLE );
			}

		//---------------------------------------------------------------------
		void SessionOpen( const char * pchFile, bool quick ) { 
		
			//----- Get session settings from file (if loading) or from default (if new).
			if( pchFile==NULL ) {		//-- New session - create default values 				
				m_oSession.NewSession( g_oCfgs.m_oSettings );
				SetScanMode( g_oCfgs.m_oSettings.m_bScrollView ? 1 : 2 );
				m_bTimeSep = g_oCfgs.m_oSettings.m_bTimeSep;
				}
			else {						//-- load session
				if( ! m_oSession.LoadFile(pchFile, &m_iScanMode, &m_bTimeSep, m_rctDisp.right-m_rctDisp.left-SCOPE_XPOS ) ) {
					MessageBox( m_hDlg, "Error loading session file", pchFile, MB_OK);
					DestroyWindow( GetParent( m_hDlg ) );
					return;
					}
				SetScanMode( m_iScanMode );
				m_cLongName.Set( pchFile );
				SetWindowText( GetParent(m_hDlg), pchFile );
				}
				
			//----- Present config dialogue to get options.
			m_oSession.m_bQuick = quick;
			if( IDOK != DialogBoxParam( GetInstance(), MAKEINTRESOURCE(IDD_DLGCONFIGSESSION), GetParent(m_hDlg), (DLGPROC)ConfigSessionDlgProc, (LPARAM)&m_oSession ) ) {
				DestroyWindow( GetParent( m_hDlg ) );	// if cancel of new session config, delete the session view
				return;
				}

			//-------------------- Setup Session
			//-- setup the hardware pointers
			m_oSession.LoadConfig( );
			
			//-- set the range for the slider control 
			int iTimeDivPos = max( MINSPEED, min( MAXSPEED, m_oSession.m_iAcqSampPerPixel));
			SendMessage( m_sliderTimeDiv.GetHandle(), TBM_SETRANGE, (WPARAM) FALSE, (LPARAM) MAKELONG(MINSPEED,MAXSPEED) );
			SendMessage( m_sliderTimeDiv.GetHandle(), TBM_SETPOS, (WPARAM) TRUE, (LPARAM) iTimeDivPos  );
			
			//-- count the channels and load the protocols, 
			m_oSession.LoadProtocols( );
			int iGen = m_oSession.m_poCfg->ActiveGenChannels();
			bool bDigGen = m_oSession.m_poCfg->HasProtocolDigGen();
			int iAcq = m_oSession.m_poCfg->ActiveAcqChannels();
			bool bDigAcq = m_oSession.m_poCfg->DigAcqCount()>0;

			//-- if generate only make some changes 
			if( iAcq==0 && ! bDigAcq ) { 
				m_btnPlay.SetVisible(false);
				m_btnRecord.SetVisible(false);
				m_btnSynch.SetVisible(false);
				m_btnStop.SetVisible(false);
				m_btnTimeSep.SetVisible(false);
				m_btnScanMode[0].SetVisible(false);
				m_btnScanMode[1].SetVisible(false);
				m_btnQub.SetVisible(false);
				}
				
			if( iAcq==0 ) { 
				m_btnDrawMode[0].SetVisible(false);
				m_btnDrawMode[1].SetVisible(false);
				m_btnDrawMode[2].SetVisible(false);
				}
				
			//-- if acq only - remove the synch and GenAll button 
			if( iGen==0 && ! bDigGen ) {
				m_btnSynch.SetVisible(false);
				m_btnGenAll.SetVisible(false);
				m_btnGenAllStop.SetVisible(false);
				}

			//-- setup the panels arrays
			if( pchFile==NULL ) {
				int iExtraHgt = ANAGENPANEL_HEIGHT*iGen + (bDigGen?DIGGENPANEL_HEIGHT:0) + (bDigAcq?DIGACQPANEL_HEIGHT:0);
				int iPnlHgt = max(2,(m_rctDisp.bottom-m_rctDisp.top-iExtraHgt-1) / max(1,iAcq));
				int iPnlWid = m_rctDisp.right-m_rctDisp.left-SCOPE_XPOS;
				m_oSession.CreatePanels( iPnlHgt, iPnlWid, m_rctDisp.top, m_rctDisp.left+SCOPE_XPOS+1 );
				}
			else 
				m_oSession.SetupPanels( );

			//-- Create bitmaps for buffering display 
			HDC hdc = GetDC( m_hDlg );
			m_oSession.SetupBitmaps(hdc);
			ReleaseDC( m_hDlg, hdc );
			
			//-- Setup the manual digital output display
			for( int iDigout=0; iDigout<CONFIG_DOROWS; ++iDigout ) {
				m_chkDigout[iDigout].SetCheck(false);
				m_chkDigout[iDigout].SetVisible(false);
				}
				
			if( m_oSession.IsDigoutControl() ) {
				m_aDigoutControl = new char[DIGOUTCTL_BUFSIZE];
				//-- bytes 0,1,2 (3 unused)
				m_aDigoutControl[0]=32;		// TODO - determine if card has 8 or 32 lines on port0
				m_aDigoutControl[1]=0;		// No port initialization
				m_aDigoutControl[2]=0;		// not cancelled
				
				//-- bytes 4..7 = auto reset flags
				unsigned int iBits=0;		
				for( int iDigout=0; iDigout<CONFIG_DOROWS; ++iDigout ) 
					if( m_oSession.m_poCfg->m_aiDigMode[iDigout]==DOMODE_AUTO )
						iBits |= (1<<iDigout);
				((int*)m_aDigoutControl)[1]=iBits;	
				
				//-- bytes[8..11] Initial value and communication area for writes
				((int*)m_aDigoutControl)[2]=0;	

				//-- bytes[12..15] Last value written 
				((int*)m_aDigoutControl)[3]=0;	

				//-- bytes[16..?] port name string 				
				// TODO - can reduce sring size using "/line0:4" format for ranges.
				m_aDigoutControl[16]=0;
				int iBufLen=0;
				for( int iDigout=0; iDigout<CONFIG_DOROWS; ++iDigout ) 
					if( m_oSession.m_poCfg->m_aiDigMode[iDigout]!=DOMODE_INPUT 
							&& m_oSession.m_poCfg->m_aiDigMode[iDigout]!=DOMODE_DISABLED
							&& iBufLen+32<DIGOUTCTL_BUFSIZE ) {
						if( iBufLen>0 ) 
							m_aDigoutControl[16+iBufLen++]=',';
						sprintf( m_aDigoutControl+16+iBufLen, "%s/port0/line%d", m_oSession.m_poCard->m_achName, iDigout );
						iBufLen=strlen(m_aDigoutControl+16);
						}
									
				//-- button display string
				m_oSession.m_iDigoutStatus=0;
				if( CreateThread( NULL, 0, DigitalOutputPoller, m_aDigoutControl, 0, NULL) == NULL )
					ErrorMessage( IDS_DIGOUTTHREADCREATE );
				else
					for( int iDigout=0; iDigout<CONFIG_DOROWS; ++iDigout ) {
						if( m_oSession.IsDigoutControl( iDigout ) ) {
							m_chkDigout[iDigout].SetVisible( true );
							m_chkDigout[iDigout].SetText( m_oSession.m_poCfg->m_acDigName[iDigout].ConstBuffer() );
							m_oSession.m_iDigoutStatus += (1<<iDigout);
							}
						}
				}

			//-- setup the feedback controller
			m_oSession.CreateFeedback();

			//-- prepare to display 			
			if( pchFile==NULL )
				m_oSession.SetPanelPositions( m_rctDisp.top+1 );

			CreatePanelExtras( true, -1 );
			//Unnecc...ShowControls();						//-- Generation panel extras
			
			OnExitSizing(NULL);
			InvalidateRect( m_hDlg, NULL /*&m_rctDisp*/, false );
			m_bSetup=true;
			}

		//---------------------------------------------------------------------
		void OnPaint() {
			if( ! m_bSetup )
				return;
			
			PAINTSTRUCT ps;
			HDC hdc = BeginPaint(m_hDlg, &ps);
			
			//----- Manual Drawing 
			if( m_aiCapture[0]==NOCAPTURE ) {
				if( ps.rcPaint.top < m_rctDisp.top
						|| ps.rcPaint.left < m_rctDisp.left+SCOPE_XPOS
						|| ps.rcPaint.bottom > m_rctDisp.bottom
						|| ps.rcPaint.right > m_rctDisp.right ) {
					m_oSession.DrawBgr( hdc, m_iPage );
					DrawExtras( hdc );		//-- Draw the 'Extras' - special controls
					}
				m_oSession.DrawFgr( hdc, m_iScanMode );
				FocusRect( -1, -1, -1, -1, hdc );
				}
			
			//----- Show dragging line when sizing
			if( m_aiCapture[0]!=NOCAPTURE ) {
				CSessionPanel * pPnl = m_oSession.Panel(m_aiCapture[5]);
				FocusRect(	m_aiCapture[0]==CAP_LFT ? m_aiCapture[3] : pPnl->m_iLft-SCOPE_XPOS,
							m_aiCapture[0]==CAP_TOP ? m_aiCapture[4] : pPnl->m_iTop, 
							m_aiCapture[0]==CAP_RGT ? m_aiCapture[3] : pPnl->m_iLft+pPnl->m_iWid,
							m_aiCapture[0]==CAP_BTM ? m_aiCapture[4] : pPnl->m_iTop+pPnl->m_iHgt, 
							hdc );
				}

			
			//----- Done
			EndPaint(m_hDlg, &ps);
			}
			
		//-- create or recreate the 'extra' controls for the panels
		// The acq panel has a series of 'Extras', the gen panel has a series of 'controls'.
		// The first time this is called, all are setup.  subsequent calls only setup the 
		// acq since it is in response to a resize which does not affest the gen.
		// 
		void CreatePanelExtras( bool bFirst, int iPanel ) { 
			CSessionPanel * pPnl;

			//-- Acquisition panel extras
			for( int iPnl=0; iPnl<m_oSession.m_aPanels.Count(); ++iPnl ) {
				pPnl = m_oSession.Panel(iPnl);

				if( ( iPanel==-1 || iPanel==iPnl ) && pPnl->m_iPage>=0 ) {

					int iLeft = pPnl->m_iLft-SCOPE_XPOS;
					int iRight = pPnl->m_iLft+pPnl->m_iWid;
					if( pPnl->m_iChannelType==CT_AI ) { 
						//-- Acquisition : Y axis Scroll bar
						DestroyExtra(ID_VBAR+iPnl);
						int iX =pPnl->m_iLft - YSCROLL_WIDTH;
						pPnl->pVBar = new CDlgMiniScroll( ID_VBAR+iPnl, m_hDlg, iX, pPnl->m_iTop, iX+YSCROLL_WIDTH-1, pPnl->m_iTop+pPnl->m_iHgt, 
										pPnl->m_dScaling*pPnl->m_dMin, pPnl->m_dScaling*pPnl->m_dMax,
										pPnl->m_dScaling*pPnl->m_dMinDisp, pPnl->m_dScaling*pPnl->m_dMaxDisp );
						m_aExtra.AddItem( pPnl->pVBar );

						//-- title text
						DestroyExtra(ID_TITLE+iPnl);
						m_aExtra.AddItem( new CDlgVertText( ID_TITLE+iPnl, m_hDlg, iLeft+1, pPnl->m_iTop+1, 
										iLeft+TITLE_WIDTH, pPnl->m_iTop+pPnl->m_iHgt-1, pPnl->m_cName.Buffer(), 12 ) );

						//-- units text 									
						DestroyExtra(ID_UNITS+iPnl);
						m_aExtra.AddItem( new CDlgVertText( ID_UNITS+iPnl, m_hDlg, iLeft+TITLE_WIDTH+1, pPnl->m_iTop+1, 
										iLeft+TITLE_WIDTH+UNITS_WIDTH, pPnl->m_iTop+pPnl->m_iHgt-1, pPnl->m_cUnits.Buffer(), 12 ) );
						}

					else if( pPnl->m_iChannelType==CT_AO ) { 
						//----- Generation - Edit controls, play/loop/stop buttons.
						CProtocol * pPro = &(m_oSession.m_aProtocol[pPnl->m_iChannelIndex]);
						CProtocolSegment * pSeg;
						int iOffset=iLeft+SCOPE_XPOS+UPDATE_WIDTH+2;
						int iOffsetMax = iLeft+pPnl->m_iWid-UPDATE_WIDTH*3-EDIT_WIDTH;
						int iPnlID=ID_GENCTL+(1+pPnl->m_iChannelIndex)*ID_GENCTLX;
						int iSegID;

						//----- For resizing - remove prior controls
						for( int iInc=0; iInc<=12; ++iInc ) 
							DestroyControl(iPnlID+iInc);
						for( int iPnlSeg=0; iPnlSeg<pPro->m_aSeg.Count(); iPnlSeg++ ) { 
							iSegID=iPnlID+(1+iPnlSeg)*ID_GENCTLPNLX;
							for( int iInc=0; iInc<=8; ++iInc ) 
								DestroyControl( iSegID+iInc );
							}
						
						//-- title / units 
						CString cTitle( pPnl->m_cName );
						cTitle += " ";
						cTitle += pPnl->m_cUnits.Buffer();
						CreateLabel( iLeft+1, pPnl->m_iTop+1, SCOPE_XPOS, 16, iPnlID+0, NULL, g_fntSS8 )->SetText(cTitle.Buffer());

						//-- Iteration Count 								
						if( iOffset<iOffsetMax ) { 
							CreateLabel( iOffset, pPnl->m_iTop+1, 25, 16, iPnlID+1, NULL, g_fntSS8 )->SetText("Iter:");
							iOffset+=27;

							CreateEdit( iOffset, pPnl->m_iTop+1, EDIT_WIDTH-5, 16, iPnlID+2 )->SetTextInt( pPro->m_iIterations ); 
							iOffset += EDIT_WIDTH;
							}

						//-- per segment controls
						bool bModifiable=false;
						for( int iPnlSeg=0; iPnlSeg<pPro->m_aSeg.Count(); iPnlSeg++ ) 
							if( pPro->Segment( iPnlSeg )->m_bModifiable ) { 
								bModifiable=true;
								break;
								}

						//-- if any modifiable - add an iteration count and update button
						if( bModifiable )
							pPnl->m_pUpdate = CreateButton( iLeft+SCOPE_XPOS, pPnl->m_iTop+1, UPDATE_WIDTH, UPDATE_WIDTH, iPnlID+3, g_hbUpdate[1], "Update protocol" );
						else
							pPnl->m_pUpdate = NULL;

						for( int iPnlSeg=0; iPnlSeg<pPro->m_aSeg.Count(); iPnlSeg++ ) { 
							pSeg = pPro->Segment( iPnlSeg );
							if( pSeg->m_bModifiable ) { 
								//----- for modifiable segments : set base ID
								iSegID=iPnlID+(1+iPnlSeg)*ID_GENCTLPNLX;
							
								//-- add segment label to the panel
								CreateLabel( iOffset, pPnl->m_iTop+1, 25, 16, iSegID+1, NULL, g_fntSS8 )->SetTextInt( iPnlSeg+1, " %d:");
								if( (iOffset+=20)>=iOffsetMax )
									break;
								
								//-- add duration edit to the panel
								CreateEdit( iOffset, pPnl->m_iTop+1, EDIT_WIDTH-5, 16, iSegID+2)->SetText( pSeg->m_cDuration.ConstBuffer() ); 
								if( (iOffset+=EDIT_WIDTH)>=iOffsetMax )
									break;
								
								//-- might be a f(u), add u to the edit.
								if( strchr(pSeg->m_cFunction.Buffer(),'u') != NULL ) {
									CreateLabel( iOffset, pPnl->m_iTop+1, 30, 16, iSegID+3, NULL, g_fntSS8 )->SetText(" u");
									if( (iOffset+=20)>=iOffsetMax )
										break;
									CreateEdit( iOffset, pPnl->m_iTop+1, EDIT_WIDTH-5, 16, iSegID+4)->SetTextDouble(pSeg->GetV0());
									if( (iOffset+=EDIT_WIDTH)>=iOffsetMax )
										break;
									if( pSeg->m_bV0Range ) { 
										CreateEdit( iOffset, pPnl->m_iTop+1, EDIT_WIDTH-5, 16, iSegID+7)->SetTextDouble(pSeg->GetV0Final());
										if( (iOffset+=EDIT_WIDTH)>=iOffsetMax )
											break;
										}
									}
									
								//-- might be a f(v), add v to the edit.								
								if( strchr(pSeg->m_cFunction.Buffer(),'v') != NULL ) {
									CreateLabel( iOffset, pPnl->m_iTop+1, 30, 16, iSegID+5, NULL, g_fntSS8 
												)->SetText(" v");
									if( (iOffset+=20)>=iOffsetMax )
										break;
									CreateEdit( iOffset, pPnl->m_iTop+1, EDIT_WIDTH-5, 16, iSegID+6)->SetTextDouble(pSeg->GetV1());
									if( (iOffset+=EDIT_WIDTH)>=iOffsetMax )
										break;
									if( pSeg->m_bV1Range ) { 
										CreateEdit( iOffset, pPnl->m_iTop+1, EDIT_WIDTH-5, 16, iSegID+8)->SetTextDouble(pSeg->GetV1Final());
										if( (iOffset+=EDIT_WIDTH)>=iOffsetMax )
											break;
										}
									}
								}
							}
							
						//-- Feedback control values( PID specific now, more generic methodology if other feedback models are implemented).
						CPID * pPID = (CPID *)m_oSession.m_pFeedback;
						if( m_oSession.m_bFeedback && pPID != NULL && m_oSession.m_iFeedbackApply==pPnl->m_iChannelIndex ) { 
							//-- Add PID(p) setting 
							CreateLabel( iOffset, pPnl->m_iTop+1, 25, 16, iPnlID+7, NULL, g_fntSS8 )->SetText("P:");
							iOffset+=20;
							CreateEdit( iOffset, pPnl->m_iTop+1, EDIT_WIDTH-5, 16, iPnlID+8 )->SetTextDouble( pPID->m_dKp );
							if( (iOffset+=EDIT_WIDTH)>=iOffsetMax )
								break;
								
							//-- Add PID(i) setting 
							CreateLabel( iOffset, pPnl->m_iTop+1, 25, 16, iPnlID+9, NULL, g_fntSS8 )->SetText("I:");
							iOffset+=20;
							CreateEdit( iOffset, pPnl->m_iTop+1, EDIT_WIDTH-5, 16, iPnlID+10 )->SetTextDouble( pPID->m_dKi );
							if( (iOffset+=EDIT_WIDTH)>=iOffsetMax )
								break;
								
							//-- Add PID(d) setting 
							CreateLabel( iOffset, pPnl->m_iTop+1, 25, 16, iPnlID+11, NULL, g_fntSS8 )->SetText("D:");
							iOffset+=20;
							CreateEdit( iOffset, pPnl->m_iTop+1, EDIT_WIDTH-5, 16, iPnlID+12 )->SetTextDouble( pPID->m_dKd );
							if( (iOffset+=EDIT_WIDTH)>=iOffsetMax )
								break;
							}

						//-- Add button for [Play][Play continuous][Stop]
						pPnl->m_pPlay = CreateButton( iRight-5-UPDATE_WIDTH*3, pPnl->m_iTop+1, UPDATE_WIDTH, UPDATE_WIDTH, iPnlID+4, g_hbPlay[2], "Generate once" );
						pPnl->m_pLoop = CreateButton( iRight-5-UPDATE_WIDTH*2, pPnl->m_iTop+1, UPDATE_WIDTH, UPDATE_WIDTH, iPnlID+5, g_hbLoop[2], "Generate continuously" );
						pPnl->m_pStop = CreateButton( iRight-5-UPDATE_WIDTH*1, pPnl->m_iTop+1, UPDATE_WIDTH, UPDATE_WIDTH, iPnlID+6, g_hbStop[1], "Stop generation" );
						}

					else if( pPnl->m_iChannelType==CT_DI ) { 
						int iPnlID=ID_DIGACQCTL;
						DestroyControl( iPnlID+0 );
						CreateLabel( iLeft+1, pPnl->m_iTop+1, SCOPE_XPOS, 16, iPnlID+0, NULL, g_fntSS8)->SetText("Dig Acq");
						}

					else if( pPnl->m_iChannelType==CT_DO ) { 
						int iPnlID=ID_DIGGENCTL;

						DestroyControl( iPnlID+0 );
						DestroyControl( iPnlID+4 );
						DestroyControl( iPnlID+5 );
						DestroyControl( iPnlID+6 );
						
						CreateLabel( iLeft+1, pPnl->m_iTop+1, SCOPE_XPOS, 16, iPnlID+0, NULL, g_fntSS8)->SetText("Dig Gen");

						//-- Add button for [Play][Play continuous][Stop]
						pPnl->m_pPlay = CreateButton( iRight-5-UPDATE_WIDTH*3, pPnl->m_iTop+1, UPDATE_WIDTH, UPDATE_WIDTH, iPnlID+4, g_hbPlay[2], "Generate once" );
						pPnl->m_pLoop = CreateButton( iRight-5-UPDATE_WIDTH*2, pPnl->m_iTop+1, UPDATE_WIDTH, UPDATE_WIDTH, iPnlID+5, g_hbLoop[2], "Generate continuously" );
						pPnl->m_pStop = CreateButton( iRight-5-UPDATE_WIDTH*1, pPnl->m_iTop+1, UPDATE_WIDTH, UPDATE_WIDTH, iPnlID+6, g_hbStop[1], "Stop generation" );
						}
					}
				}
			}
			
		void SetGenButtonBitmaps( int iAcqMode, int iGenMode ) { 
			for( int iPnl=0; iPnl<m_oSession.m_aPanels.Count(); ++iPnl ) { 
				CSessionPanel * pPnl = m_oSession.Panel(iPnl);
				if( m_oSession.m_iPage==pPnl->m_iPage ) { 
					if( pPnl->m_iChannelType==CT_AO || pPnl->m_iChannelType==CT_DO ) { 
						pPnl->m_pPlay->SetButtonBitmap( g_hbPlay[ m_iAcqMode!=MODE_SYNCH && m_iGenMode==MODE_IDLE ? 2 : 1 ] );
						pPnl->m_pLoop->SetButtonBitmap( g_hbLoop[ m_iAcqMode!=MODE_SYNCH && m_iGenMode==MODE_IDLE ? 2 : 1 ] );
						//pPnl->m_pStop->SetButtonBitmap( g_hbStop[ m_iAcqMode!=MODE_SYNCH && (m_iGenMode|MODE_GENCONT)==(pPnl->m_iChannelIndex|MODE_GENCONT) ? 2 : 1 ] );
						pPnl->m_pStop->SetButtonBitmap( g_hbStop[ 2 ] );
						}
					}
				}
			}
			
		//-- Process message to update data display for new acquired data 
		void OnDataUpdate( bool bRecord ) { 
			CSession * pSes = &m_oSession;
			bool bUpdate = false;				// if any updates are made 
			int iSesPos = pSes->m_iQueuePos;	// check position now, polling thread can change it during update
			
			//-- for each completed pixel
			HPEN penWhite = CreatePen( 0, 0, RGB(255,255,255) );	// for clearing columns
			HPEN penSep = CreatePen( 0, 0, RGB(64,128,192) );		// for time sep every 1k samples
			HPEN penToUse;
				
			while( ! pSes->m_pAcqInit->m_bCancelled
					&& iSesPos > pSes->m_iAcqSampPerPixel + ( iSesPos>=m_iQuePos ? m_iQuePos : 0 ) 
					) { 

				//-- clear a column ( white or time sep color ) 
				if( m_bTimeSep && m_iTimeBar+m_iTimeSep<m_iElapsed ) {
					penToUse=penSep;
					m_iTimeBar=m_iTimeSep*int(m_iElapsed/m_iTimeSep);
					}
				else 
					penToUse=penWhite;

				//-- For each visible panel
				for( int iPanel=0; iPanel<m_oSession.m_aPanels.Count(); ++iPanel ) {
					CSessionPanel * pPnl = m_oSession.Panel(iPanel);

					//-- If visible then  update display 
					if( pPnl->m_iPage>=0 ) { 
						
						//----- Analog acquisition 
						if( pPnl->m_iChannelType==CT_AI )
							pPnl->AIUpdate( pSes->m_iAcqSampPerPixel, pSes->m_adQueue, 
								pSes->m_poAnaAcqTask->m_iAcqChannels, iPanel, pSes->m_iQueueSize, m_iQuePos, bRecord, penToUse, m_iScanMode, m_iDrawMode );
						
						//----- Digital acquisition 
						if( pPnl->m_iChannelType==CT_DI ) 
							pPnl->DIUpdate( pSes->m_iAcqSampPerPixel, pSes->m_aiDigQueue,
								pSes->m_iQueueSize, m_iQuePos, pSes->m_iDigAcqLines, pSes->m_aiDigAcqMap, m_iScanMode, m_iDrawMode );

						} 
					}

				//----- 
				m_iQuePos=(m_iQuePos+pSes->m_iAcqSampPerPixel)%pSes->m_iQueueSize;
				bUpdate=true;
				m_iElapsed += pSes->m_iAcqSampPerPixel ;
				}
				
			DeleteObject( penWhite );
			DeleteObject( penSep );
			
			SetElapsed( m_iElapsed );
			
			if( bUpdate ) { 
				RECT r = m_rctDisp;
				r.left += SCOPE_XPOS;
				r.bottom = m_rctDisp.bottom;
				InvalidateRect( m_hDlg, &r, false );
				}
			}

		//-----------------------------------------------------------------------------			
		// assumes ai panels are first in list;  assumes visible since vbar changed.
		void OnChangeVBar( int iPanel ) { 			
			CSessionPanel * pPnl = m_oSession.Panel(iPanel);
			pPnl->m_dMinDisp = ((CDlgMiniScroll *)m_pCapture)->GetLowVal() / pPnl->m_dScaling;
			pPnl->m_dMaxDisp = ((CDlgMiniScroll *)m_pCapture)->GetHighVal() / pPnl->m_dScaling;

			//RECT rct;
			//SetRect( &rct, pPnl->m_iLft-SCOPE_XPOS, pPnl->m_iTop, pPnl->m_iLft, pPnl->m_iTop+pPnl->m_iHgt );
			//InvalidateRect( m_hDlg, &rct, TRUE );
			
			InvalidateRect( m_hDlg, NULL, true );
			}

		void OnChangeTimeDiv() { 			
			int iPos = SendMessage( m_sliderTimeDiv.GetHandle(), TBM_GETPOS, 0, 0 );
			m_oSession.m_iAcqSampPerPixel = iPos;
			}

		//---------------------------------------------------------------------
		void OnExit() { 
			// Stop acquisition
			if( m_iAcqMode!=MODE_IDLE ) { 
				m_oSession.Stop();
				m_oSession.m_pAcqInit->m_pSession=NULL;
				} 
			// stop generation
			if( m_iGenMode!=MODE_IDLE ) { 
				m_oSession.GenStop();
				m_oSession.m_pGenInit->m_pSession=NULL;
				} 
			//stop digital output poller
			if( m_aDigoutControl!=NULL ) 
				m_aDigoutControl[2]=1;
			}

		void OnBtnQub() { 			
			if( m_iAcqMode!=MODE_IDLE || m_iGenMode!=MODE_IDLE ) 
				return;
			m_oSession.m_oQDF.WriteDataNotes();
			m_oSession.m_oQDF.Close();
			CString cStart("start ");		// "\"\\Program files\\qub\\qub.exe\" "
			cStart.Append('"');
			cStart.Append("sending to qub...");
			cStart.Append('"');
			cStart.Append(' ');
			cStart.Append('"');
			cStart.Append( m_oSession.m_cAcqFile.Buffer() );
			cStart.Append('"');
			system( cStart.Buffer() );
			DestroyWindow( GetParent( m_hDlg ) );
			}

		bool OnBtnSave() { 			
			if( m_iAcqMode != MODE_IDLE || m_iGenMode!=MODE_IDLE ) 
				return false;
			if( m_cLongName.Length()==0 ) { 
				char bufName[_MAX_PATH+1]="";
				//-- select the file to save
				if( ! GetSaveFileName(bufName, 
								"Session Files (*.qsn)\0*.qsn\0All Files (*.*)\0*.*\0\0",
								"qsn", 
								GetParent(m_hDlg), NULL,
								& g_oCfgs.m_oSettings.m_cSessionPath ) ) 
					return false;
				m_cLongName.Set(bufName);
				}
			if( ! m_oSession.SaveFile( m_cLongName.Buffer(), m_iScanMode, m_bTimeSep  ) ) {
				m_cLongName.Set("");
				MessageBox( m_hDlg, "Error saving session file", "QUBIO", MB_OK );
				return false;
				}
			SetWindowText( GetParent(m_hDlg), m_cLongName.Buffer() );

			return true;
			}

		void OnBtnSaveClose() { 			
			if( OnBtnSave() )
				DestroyWindow( GetParent( m_hDlg ) );
			}

		void OnBtnConfig() { 			
			DialogBoxParam( GetInstance(), MAKEINTRESOURCE(IDD_DLGCONFIGSESSION), GetParent(m_hDlg), (DLGPROC)ConfigSessionDlgProc, (LPARAM)&m_oSession );
			}

		//---------------------------------------------------------------------
		void OnBtnDigout( int iDigout, bool bChecked ) { 			
			//TODO - allow for 8 or 32 bit ports 
			//-- construct output status byte 
			unsigned char chOut=0;

			// Write Digout status 
			char bufLabel[64];
			if( m_oSession.m_poCfg->m_aiDigMode[iDigout]==DOMODE_AUTO ) 
				sprintf( bufLabel, "%s", m_oSession.m_poCfg->m_acDigName[iDigout].Buffer() );
			else 
				sprintf( bufLabel, "%s %s", m_oSession.m_poCfg->m_acDigName[iDigout].Buffer(), bChecked ? "On" : "Off" );
			
			//= "[- - - - - - - -]     ";
			for( int i=0; i<CONFIG_DOROWS; ++i ) 
				if( m_chkDigout[i].GetCheck() )
					chOut += 1<<i;
					
			int iStatus = bChecked ? 1 : 0;
			if( m_oSession.m_poCfg->m_aiDigMode[iDigout]==DOMODE_AUTO ) {
				m_chkDigout[iDigout].SetCheck( false );
				iStatus=2;
				}
		
			m_oSession.m_oQDF.AddNote( m_oSession.m_iAcqPos, bufLabel, "" );

			m_oSession.RecordDigout( iDigout, iStatus );

			//-- update status int that Digout poller will discover and send 
			((int*)m_aDigoutControl)[2]=int(chOut);
			}

		//---------------------------------------------------------------------
		void OnBtnPlay() { 			
			//-- switch from Idle to Playing
			if( m_iAcqMode == MODE_IDLE ) { 
				m_iQuePos=0;
				if( m_oSession.Play(ACQ_QUE_SIZE) ) 
					SetModes( MODE_ACQPLAY, MODE_NULL);
				}
			//-- Switch from Recording to playing 
			if( m_iAcqMode == MODE_ACQRECORD ) { 
				m_oSession.EndRecording();
				SetModes( MODE_ACQPLAY, MODE_NULL );
				}
			}
		
		void OnBtnRecord() { 			
			//-- switch from Idle to Recording
			if( m_iAcqMode == MODE_IDLE ) { 
				m_iQuePos=0;
				if( m_oSession.Record(ACQ_QUE_SIZE) ) 
					SetModes( MODE_ACQRECORD, MODE_NULL );
				}
			//-- Switch from playing to Recording 
			if( m_iAcqMode == MODE_ACQPLAY ) { 
				m_oSession.StartRecording();
				SetModes( MODE_ACQRECORD, MODE_NULL );
				}
			}
		
		void OnBtnStop() { 			
			if( ( m_iAcqMode == MODE_ACQPLAY || m_iAcqMode == MODE_ACQRECORD )
				&& m_iGenMode != MODE_IDLE
				&& m_oSession.m_bFeedback ) {
				AddMessage( REPTYPE_ERROR, "With feedback active, generation must be stopped prior to ending acquisition" );
				return;
				}

			if( m_iAcqMode != MODE_IDLE && m_oSession.Stop() )
				SetModes( MODE_IDLE, MODE_NULL );
			}

		//---------------------------------------------------------------------
		void OnBtnSynch() { 			
			if( m_iAcqMode==MODE_IDLE && m_iGenMode == MODE_IDLE ) {
				m_iQuePos=0;
				if( m_oSession.Synch( ACQ_QUE_SIZE ) )
					SetModes( MODE_SYNCH, MODE_NULL );
				}
			}

		//---------------------------------------------------------------------
		void OnBtnGenAll( ) { 
			if( m_iAcqMode == MODE_SYNCH || m_iGenMode != MODE_IDLE ) { 
				AddMessage( REPTYPE_NOTE, "A generation task is already active" );
				return;
				}
			if( ( m_iAcqMode != MODE_ACQPLAY && m_iAcqMode != MODE_ACQRECORD )
				&& m_oSession.m_bFeedback ) {
				AddMessage( REPTYPE_ERROR, "With feedback active, acquisition must be started prior to generation.");
				return;
				}
			if( m_oSession.Generate( -1, true ) ) 
				SetModes( MODE_NULL,  MODE_GENALL );
			}

		void OnBtnGenAllStop( ) { 
			if( m_iGenMode != MODE_GENALL ) 
				return;
			if( m_oSession.GenStop( ) ) 
				SetModes( MODE_NULL, MODE_IDLE );
			}

		//---------------------------------------------------------------------
		void OnBtnGen( int iPro, bool bCont ) { 
			if( m_iAcqMode == MODE_SYNCH || m_iGenMode != MODE_IDLE ) { 
				AddMessage( REPTYPE_NOTE, "A generation task is already active" );
				return;
				}
			if( ( m_iAcqMode != MODE_ACQPLAY && m_iAcqMode != MODE_ACQRECORD )
				&& m_oSession.m_bFeedback ) {
				AddMessage( REPTYPE_ERROR, "With feedback active, acquisition must be started prior to generation.");
				return;
				}
			if( m_oSession.Generate( iPro, bCont ) ) 
				SetModes( MODE_NULL,  iPro | (bCont?MODE_GENCONT:MODE_GENONCE) );
			}

		//---------------------------------------------------------------------
		void OnBtnGenStop( int iPro ) { 
			if( ( (m_iGenMode&0xFF000)!=MODE_GENCONT && (m_iGenMode&0xFF000)!=MODE_GENONCE ) 
				|| ((m_iGenMode & 0xFFF) != iPro ) )
				return;
		
			if( m_oSession.GenStop( ) ) 
				SetModes( MODE_NULL, MODE_IDLE );
			}
			
		//---------------------------------------------------------------------
		void OnBtnDigGen( int iLine, bool bCont ) {			//-- iLine=-1 for all
			if( m_iAcqMode == MODE_SYNCH || m_iGenMode != MODE_IDLE ) { 
				AddMessage( REPTYPE_NOTE, "A generation task is already active" );
				return;
				}
			if( ( m_iAcqMode != MODE_ACQPLAY && m_iAcqMode != MODE_ACQRECORD )
				&& m_oSession.m_bFeedback ) {
				AddMessage( REPTYPE_ERROR, "With feedback active, acquisition must be started prior to generation.");
				return;
				}
			if( m_oSession.Generate( iLine, bCont ) ) 
				SetModes( MODE_NULL,  iLine | (bCont?MODE_GENCONT:0) );
			}

		//---------------------------------------------------------------------
		void OnBtnDigGenStop( int iLine ) {		//-- iLine=-1 for all
			if( (m_iGenMode|MODE_GENCONT)!=(iLine|MODE_GENCONT) )
				return;
			if( m_oSession.GenStop( ) ) 
				SetModes( MODE_NULL, MODE_IDLE );
			}

		//---------------------------------------------------------------------
		void OnBtnGenUpdate( int iChn ) { 
			//-- change the update button icon back to gray			
			CSessionPanel * pPnl = m_oSession.Panel(m_oSession.m_aiGenChannelPanel[iChn]);
			pPnl->m_pUpdate->SetButtonBitmap( g_hbUpdate[1] );

			//-- get values from the edits
			char buf[100];
			CProtocolSegment * pSeg;
			int iPnlID=ID_GENCTL+(1+pPnl->m_iChannelIndex)*ID_GENCTLX;
			int iSegID;

			GetWindowText( GetDlgItem( m_hDlg, iPnlID+2 ), buf, 100 );
			m_oSession.m_aProtocol[iChn].m_iIterations = atoi( buf );
			
			for( int iPnlSeg=0; iPnlSeg<m_oSession.m_aProtocol[iChn].m_aSeg.Count(); iPnlSeg++ ) { 
				pSeg = m_oSession.m_aProtocol[iChn].Segment( iPnlSeg );
				if( pSeg->m_bModifiable ) { 
					//-- update the duration, u, v parameters
					iSegID=iPnlID+(1+iPnlSeg)*ID_GENCTLPNLX;
					
					if( GetWindowText( GetDlgItem( m_hDlg, iSegID+2 ), buf, 100 ) >= 0 ) 
						pSeg->SetDuration( buf );
						
					if( GetWindowText( GetDlgItem( m_hDlg, iSegID+4 ), buf, 100 ) >=0 ) /*strchr(pSeg->m_cFunction.Buffer(),'u') != NULL && */
						pSeg->SetV0( atof(buf) );
					if( GetWindowText( GetDlgItem( m_hDlg, iSegID+7 ), buf, 100 ) >=0 ) 
						pSeg->SetV0Final( atof(buf) );

					if( GetWindowText( GetDlgItem( m_hDlg, iSegID+6 ), buf, 100 ) >=0 ) /* strchr(pSeg->m_cFunction.Buffer(),'v') != NULL */
						pSeg->SetV1( atof(buf) );
					if( GetWindowText( GetDlgItem( m_hDlg, iSegID+8 ), buf, 100 ) >=0 ) 
						pSeg->SetV1Final( atof(buf) );
					}
				}
				
			//-- Update the feedback parameters 
			if( m_oSession.m_bFeedback 
				&& m_oSession.m_pFeedback != NULL ) { 
			
				CPID * pPID = (CPID *)(m_oSession.m_pFeedback);
				GetWindowText( GetDlgItem( m_hDlg, iPnlID+8 ), buf, 100 );
				pPID->m_dKp = atof( buf );
				GetWindowText( GetDlgItem( m_hDlg, iPnlID+10 ), buf, 100 );
				pPID->m_dKi = atof( buf );
				GetWindowText( GetDlgItem( m_hDlg, iPnlID+12 ), buf, 100 );
				pPID->m_dKd = atof( buf );
				}
					
			//-- Update the protocol internals 
			}

		//---------------------------------------------------------------------
		void OnBtnTimeSep() { 
			m_bTimeSep=m_btnTimeSep.GetCheck();
			m_iTimeBar=m_iElapsed;
			m_btnTimeSep.SetButtonBitmap( g_hbTimeSep[ m_bTimeSep ? 0 : 1 ] );
			}

		//---------------------------------------------------------------------
		void OnAcqPollerStop() { 
			// sent by poller when request to cancel or timeout has cause polling to end.
			}

		void OnGenPollerStop() { 
			SetModes( MODE_NULL, MODE_IDLE );
			}

		//---------------------------------------------------------------------
		void SetElapsed( int i ) { 
			if( i==0 ) 
				m_iTimeBar=0;
			m_iElapsed=i;
			m_lblElapsed.SetTextInt( m_iElapsed/1000, "k samp:%d" );
			}
			
		void SetDrawMode( int iMode ) { 
			m_iDrawMode=iMode;
			}

		//---------------------------------------------------------------------
		void DragInvalidate() { 
			InvalidateRect( m_hDlg, &m_rctDisp, false );
			}
			
		bool DragDown( int iX, int iY ){ //TODO - mouse parameters should be passed as short
			//-- click on draggable control item
			if( CDialog::DragDown( iX, iY ) )
				return true;
				
			//-- click on panel border to size it 
			if( m_aiCapture[0]==NOCAPTURE
					&& iX>=m_rctDisp.left-1 && iX<=m_rctDisp.right+2
					&& iY>=m_rctDisp.top-1 && iY<=m_rctDisp.bottom+2 ) {
				for( int iPnl=m_oSession.m_aPanels.Count()-1; iPnl>=0; --iPnl ) {
					CSessionPanel * p = m_oSession.Panel(iPnl);
					if( ( p->m_iChannelType==CT_AI || p->m_iChannelType==CT_DI )
							&& iY>=p->m_iTop && iY<=p->m_iTop+ICONIZE_BOX_SIZE 
							&& iX>=p->m_iLft-SCOPE_XPOS && iX<=p->m_iLft-SCOPE_XPOS+ICONIZE_BOX_SIZE )  {
						p->m_bIconized = ! p->m_bIconized;
						InvalidateRect( m_hDlg, NULL, true );
						break;
						}

					if( abs(iY-p->m_iTop)<=2 && iX>=p->m_iLft-SCOPE_XPOS && iX<=p->m_iLft+p->m_iWid )  
						m_aiCapture[0]=CAP_TOP;
					else if( abs(iY-(p->m_iTop+p->m_iHgt))<=2 && iX>=p->m_iLft-SCOPE_XPOS && iX<=p->m_iLft+p->m_iWid )  
						m_aiCapture[0]=CAP_BTM;
					else if( abs(iX-p->m_iLft+SCOPE_XPOS)<=2 && iY>=p->m_iTop && iY<=p->m_iTop+p->m_iHgt )  
						m_aiCapture[0]=CAP_LFT;
					else if( abs(iX-(p->m_iLft+p->m_iWid))<=2 && iY>=p->m_iTop && iY<=p->m_iTop+p->m_iHgt )  
						m_aiCapture[0]=CAP_RGT;
					else if( iX>p->m_iLft-SCOPE_XPOS && iX<p->m_iLft-YSCROLL_WIDTH && iY>=p->m_iTop && iY<=p->m_iTop+p->m_iHgt )  {
						m_aiCapture[0]=CAP_MOV;
						m_aiCapture[1]=iX-p->m_iLft;
						m_aiCapture[2]=iY-p->m_iTop;
						}

					if( m_aiCapture[0] != NOCAPTURE ) { 
						if( m_aiCapture[0]!=CAP_MOV ) { 
							m_aiCapture[1]=iX;
							m_aiCapture[2]=iY;
							m_aiCapture[3]=iX;
							m_aiCapture[4]=iY;
							}
						m_aiCapture[5]=iPnl;
						SetCapture( m_hDlg );
						DragInvalidate();
						
						return true;
						}
					}
				}
			
			
			return false;
			}
			
		void DragMove( int iX, int iY ){
			if( iX>60000 ) iX=0;
			if( iY>60000 ) iY=0;
									
			if( m_aiCapture[0]!=NOCAPTURE ) {
				if( m_aiCapture[0]==CAP_MOV )
					m_oSession.Panel(m_aiCapture[5])->MovePanel( iX-m_aiCapture[1], iY-m_aiCapture[2], m_rctDisp );
				m_aiCapture[3]=max( m_rctDisp.left+1, min( m_rctDisp.right, iX ) );
				m_aiCapture[4]=max( m_rctDisp.top+1, min( m_rctDisp.bottom, iY ) );
				DragInvalidate();
				InvalidateRect( m_hDlg, NULL, false );
				return;
				}

			if( m_pCapture->m_iID>=ID_VBAR && m_pCapture->m_iID<ID_VBAR+SESSION_MAXACQCHN ) 
				OnChangeVBar( m_pCapture->m_iID-ID_VBAR );	
							
			CDialog :: DragMove( iX, iY );
			}

		void DragUp( int iX, int iY ){
			if( iX>60000 ) iX=0;
			if( iY>60000 ) iY=0;
						
			if( m_aiCapture[0]!=NOCAPTURE ) {
				ReleaseCapture();
				
				CSessionPanel * pPnl = m_oSession.Panel(m_aiCapture[5]);
				if( m_aiCapture[0]==CAP_MOV ) {
					pPnl->MovePanel( iX-m_aiCapture[1], iY-m_aiCapture[2], m_rctDisp );
					}
				else { 
					m_aiCapture[3]=max( m_rctDisp.left+1, min( m_rctDisp.right, iX ) );
					m_aiCapture[4]=max( m_rctDisp.top+1, min( m_rctDisp.bottom, iY ) );
					pPnl->AdjustPanel( m_aiCapture[0], m_aiCapture[3], m_aiCapture[4] ); 
					HDC hdc = GetDC( m_hDlg );
					pPnl->SetupBitmap(hdc);
					ReleaseDC( m_hDlg, hdc );
					}
				
				m_aiCapture[0]=NOCAPTURE; 
				CreatePanelExtras( false, m_aiCapture[5] );
				InvalidateRect( m_hDlg, NULL, true );
				return;
				}
			
			if( m_pCapture->m_iID>=ID_VBAR && m_pCapture->m_iID<ID_VBAR+SESSION_MAXACQCHN )
				OnChangeVBar( m_pCapture->m_iID-ID_VBAR );	
			
			CDialog :: DragUp( iX,iY );
			}

		virtual void OnSize( WPARAM wparam, LPARAM * p ) {
			m_aiCapture[1] = LOWORD(*p);
			m_aiCapture[2] = HIWORD(*p);
			if ( wparam == SIZE_MAXIMIZED ) {
				m_bMaximized = true;
				OnExitSizing(p);
			} else if ( m_bMaximized && (wparam == SIZE_RESTORED) ) {
				m_bMaximized = false;
				OnExitSizing(p); // probably bad lparam to pass?
			}
		}

		virtual void OnExitSizing( LPARAM * p ) {
			if ( ! m_oSession.m_poCfg )
				return;
			m_bResizing = true;
			int iExtraHgt = ANAGENPANEL_HEIGHT*m_oSession.m_poCfg->ActiveGenChannels()
							+ (m_oSession.m_poCfg->HasProtocolDigGen() ? DIGGENPANEL_HEIGHT : 0)
							+ ((m_oSession.m_poCfg->DigAcqCount()>0) ? DIGACQPANEL_HEIGHT : 0);
			int iPrevHeight = 0; // m_rctDisp.bottom - m_rctDisp.top - iExtraHgt;
			for (int i=0; i<m_oSession.m_poCfg->ActiveAcqChannels(); ++i) // m_oSession.m_aPanels.Count(); ++i)
				iPrevHeight += m_oSession.Panel(i)->m_iHgt;
			vector<double> proportion;
			for (int i=0; i<m_oSession.m_poCfg->ActiveAcqChannels(); ++i) // m_oSession.m_aPanels.Count(); ++i)
				proportion.push_back( (iPrevHeight > 0) ? (m_oSession.Panel(i)->m_iHgt * 1.0 / iPrevHeight)
														: 0.0 );
			m_rctDisp.bottom = m_aiCapture[2] - 4;
			m_rctDisp.right = m_aiCapture[1] - 11;
			RECT rctWind;
			GetWindowRect(m_hDlg, &rctWind);
			SetWindowPos(m_hDlg, HWND_NOTOPMOST, rctWind.left, rctWind.top, m_aiCapture[1], m_aiCapture[2], SWP_NOZORDER|SWP_NOMOVE);
			GetWindowRect(m_hDlg, &rctWind);
			m_picDisp.SizeWindow( m_rctDisp.right - m_rctDisp.left, m_rctDisp.bottom - m_rctDisp.top);
			int iPanelsHgt = (m_rctDisp.bottom-m_rctDisp.top-iExtraHgt-1);
			int iPnlHgt = max(2, iPanelsHgt / max(1,m_oSession.m_poCfg->ActiveAcqChannels()));
			int iPnlWid = m_rctDisp.right-m_rctDisp.left-SCOPE_XPOS;
			int iLft = m_rctDisp.left+SCOPE_XPOS+1;
			int iTop = m_rctDisp.top;
			HDC hdc = GetDC( m_hDlg );
			int iPnl = 0;
			for (; iPnl<m_oSession.m_poCfg->ActiveAcqChannels(); ++iPnl) {
				int iOneHeight = proportion[iPnl] ? (iPanelsHgt * proportion[iPnl]) : iPnlHgt;
				CSessionPanel *pPnl = m_oSession.Panel(iPnl);
				pPnl->PlacePanel(iLft, iTop, iPnlWid, iOneHeight);
				pPnl->SetupBitmap(hdc);
				iTop += iOneHeight;
			}
			for (; iPnl<m_oSession.m_aPanels.Count(); ++iPnl) {
				int iOneHeight = ANAGENPANEL_HEIGHT; // proportion[iPnl] ? (iPanelsHgt * proportion[iPnl]) : ANAGENPANEL_HEIGHT;
				CSessionPanel *pPnl = m_oSession.Panel(iPnl);
				pPnl->PlacePanel(iLft, iTop, iPnlWid, iOneHeight);
				pPnl->SetupBitmap(hdc);
				iTop += iOneHeight;
			}
			ReleaseDC( m_hDlg, hdc );
			// m_oSession.SetPanelPositions( m_rctDisp.top+1 );
			CreatePanelExtras( true, -1 );
			InvalidateRect( m_hDlg, NULL, true );

			SetModes(m_iAcqMode, m_iGenMode);
			SetGenButtonBitmaps(m_iAcqMode, m_iGenMode);

			SendMessage(m_hDlg, WM_FINISHED_RESIZING_VWSESS, (WPARAM)0, (LPARAM)0);
		}

		void SetScanMode( int iMode ) { 			
			m_iScanMode=(iMode==2?2:1);
			m_btnScanMode[0].SetCheck( iMode!=2 );
			m_btnScanMode[1].SetCheck( iMode==2 );
			}
			
		void OpenNote() { 
			PostMessage( GetAncestor(m_hDlg, GA_ROOT),  WM_ACQNOTE_NEW, (WPARAM)(&m_oSession), (LPARAM) m_hDlg );		
			}

		void FocusRect( int iLeft, int iTop, int iRight, int iBottom, HDC hdc=NULL ) { 
			if( hdc != NULL && m_rctFocus.left!=-1 ) 
				DrawFocusRect( hdc, &m_rctFocus );

			if( iLeft != -1 ) 
				m_rctFocus.left=iLeft;
			if( iTop != -1 ) 
				m_rctFocus.top=iTop;
			if( iRight != -1 ) 
				m_rctFocus.right=iRight;
			if( iBottom != -1 ) 
				m_rctFocus.bottom=iBottom;
				
			if( hdc != NULL && m_rctFocus.left!=-1 )
				DrawFocusRect( hdc, &m_rctFocus );
			}

		void OnKeyUp( DWORD wParam, DWORD lParam ) {
			if ( wParam == VK_F6 )
				OpenNote();
			}

	};
	
//-----------------------------------------------------------------------------
BOOL CALLBACK SessionDlgProc(HWND hDlg, UINT iMsg, WPARAM wParam, LPARAM lParam) { 
	// create the CVwSession to go with the dialog 
	if( iMsg == WM_INITDIALOG ) { 
		CVwSession * pVw = new CVwSession;
		pVw->m_hDlg=hDlg;
		pVw->InitializeControls();

		SetWindowLong( GetParent(hDlg), 0, (long) pVw ) ;
		pVw->m_oSession.m_hView =hDlg;
		return TRUE;
		}
	
	// Get the CVwSession pointer for this dialog 
	CVwSession * pVwSession = (CVwSession *) GetWindowLong(  GetParent( hDlg ), 0 );

	int wmId, wmEvent;
	switch( iMsg ) { 
		case WM_SESSION_LOAD :		//-- user message for menu item to load a session
			pVwSession->SessionOpen( (char *)lParam, (bool) wParam );
			return TRUE;

		case WM_SESSION_NEW :		//-- user message for menu item to create a new session
			pVwSession->SessionOpen( NULL, false );
			return TRUE;
		
		case WM_DATA_UPDATE :		//-- generated by acquisition
			pVwSession->OnDataUpdate( wParam!=0 );
			return TRUE;

		case WM_LBUTTONDOWN :		//-- Start of drag operation			
			if( pVwSession->DragDown( LOWORD(lParam), HIWORD(lParam) ) )
				return TRUE;
			return FALSE;

		case WM_MOUSEMOVE :			//-- Mouse move during drag operation
			if( pVwSession->m_pCapture == NULL && pVwSession->m_aiCapture[0]==NOCAPTURE)
				return FALSE;
			pVwSession->DragMove( short(LOWORD(lParam)), short(HIWORD(lParam)) );
			InvalidateRect( hDlg, &(pVwSession->m_pCapture->m_rct), false );
			return TRUE;

		case WM_LBUTTONUP :			//-- end of drag operation
			if( pVwSession->m_pCapture == NULL && pVwSession->m_aiCapture[0]==NOCAPTURE)
				return FALSE;
			pVwSession->DragUp( LOWORD(lParam), HIWORD(lParam) );
			return TRUE;

		case WM_HSCROLL : 
			switch( LOWORD( wParam ) ) { 
				//case  TB_LINEUP       : OutputDebugString( "TB_LINEUP       \n" ); break;
				//case  TB_LINEDOWN     : OutputDebugString( "TB_LINEDOWN     \n" ); break;
				//case  TB_PAGEUP       : OutputDebugString( "TB_PAGEUP       \n" ); break;
				//case  TB_PAGEDOWN     : OutputDebugString( "TB_PAGEDOWN     \n" ); break;
				//case  TB_THUMBPOSITION: OutputDebugString( "TB_THUMBPOSITION\n" ); break;
				//case  TB_THUMBTRACK   : OutputDebugString( "TB_THUMBTRACK   \n" ); break;
				//case  TB_TOP          : OutputDebugString( "TB_TOP          \n" ); break;
				//case  TB_BOTTOM       : OutputDebugString( "TB_BOTTOM       \n" ); break;
				case  TB_ENDTRACK :		// note : this one always gets executed when done changing, 
										// above messages are only needed if continuous updating is desired.
					pVwSession->OnChangeTimeDiv();
				break;
				}
			break;
		
		case WM_COMMAND :
			wmId = LOWORD(wParam); 
			wmEvent = HIWORD(wParam); 
			switch( wmId ) { 
				//---- acquisition buttons
				case IDC_BTNPLAY :
					pVwSession->OnBtnPlay();
					return TRUE;
					
				case IDC_BTNSYNCH :
					pVwSession->OnBtnSynch();
					return TRUE;

				case IDC_BTNRECORD :
					pVwSession->OnBtnRecord();
					return TRUE;
					
				case IDC_BTNSTOP : 
					pVwSession->OnBtnStop();
					return TRUE;

				case IDC_BTNGENALL :
					pVwSession->OnBtnGenAll();
					return TRUE;

				case IDC_BTNGENALLSTOP :
					pVwSession->OnBtnGenAllStop();
					return TRUE;

				//----- Digital Generation buttons
				case ID_DIGGENCTL+4 :			// Play
					pVwSession->OnBtnDigGen( DIGGENCHANNELID, false );
					break;
				case ID_DIGGENCTL+5 :			// Loop
					pVwSession->OnBtnDigGen( DIGGENCHANNELID, true );
					break;
				case ID_DIGGENCTL+6 :			// Stop 
					pVwSession->OnBtnDigGenStop( DIGGENCHANNELID );
					break;

				//---- generation buttons - for each channel, once or continuous - convert wmID to channel number and continuous flag 
				// update edit changes to running generations 
				case ID_GENCTL+(1*ID_GENCTLX)+3 : 
				case ID_GENCTL+(2*ID_GENCTLX)+3 : 
				case ID_GENCTL+(3*ID_GENCTLX)+3 : 
				case ID_GENCTL+(4*ID_GENCTLX)+3 :
					pVwSession->OnBtnGenUpdate( ((wmId|1)-ID_GENCTL-ID_GENCTLX-3)/ID_GENCTLX );
					break;
				// generate once 
				case ID_GENCTL+(1*ID_GENCTLX)+4 : 
				case ID_GENCTL+(2*ID_GENCTLX)+4 : 
				case ID_GENCTL+(3*ID_GENCTLX)+4 : 
				case ID_GENCTL+(4*ID_GENCTLX)+4 : 
					pVwSession->OnBtnGen( ((wmId|1)-ID_GENCTL-ID_GENCTLX-4)/ID_GENCTLX, false );
					break;					
				// generate continuous
				case ID_GENCTL+(1*ID_GENCTLX)+5 : 
				case ID_GENCTL+(2*ID_GENCTLX)+5 : 
				case ID_GENCTL+(3*ID_GENCTLX)+5 : 
				case ID_GENCTL+(4*ID_GENCTLX)+5 : 
					pVwSession->OnBtnGen( ((wmId|1)-ID_GENCTL-ID_GENCTLX-5)/ID_GENCTLX, true );
					break;					
				// generate Stop					
				case ID_GENCTL+(1*ID_GENCTLX)+6 : 
				case ID_GENCTL+(2*ID_GENCTLX)+6 : 
				case ID_GENCTL+(3*ID_GENCTLX)+6 : 
				case ID_GENCTL+(4*ID_GENCTLX)+6 : 
					pVwSession->OnBtnGenStop( (wmId-ID_GENCTL-ID_GENCTLX-6)/ID_GENCTLX );
					break;

				//-- Digital output controls
				case IDC_DIGOUT1 : 
				case IDC_DIGOUT2 : 
				case IDC_DIGOUT3 : 
				case IDC_DIGOUT4 : 
				case IDC_DIGOUT5 : 
				case IDC_DIGOUT6 : 
				case IDC_DIGOUT7 : 
				case IDC_DIGOUT8 : 
				case IDC_DIGOUT9 : 
				case IDC_DIGOUT10: 
				case IDC_DIGOUT11: 
				case IDC_DIGOUT12: 
				case IDC_DIGOUT13: 
				case IDC_DIGOUT14: 
				case IDC_DIGOUT15: 
				case IDC_DIGOUT16: 
					pVwSession->OnBtnDigout( wmId-IDC_DIGOUT1, ( BST_CHECKED == SendMessage( GetDlgItem( hDlg, wmId ), BM_GETCHECK, 0, 0 ) ));
					return TRUE;

				//-- Add a note 
				case IDC_BTNNOTE :
					pVwSession->OpenNote();
					return TRUE;
				
				//-- config / save / close  other buttons
				case IDC_BTNCONFIG :
					pVwSession->OnBtnConfig();
					return TRUE;		

				case IDC_BTNSAVE : 
					pVwSession->OnBtnSave();
					return TRUE;

				case IDC_BTNSAVEANDCLOSE : 
					pVwSession->OnBtnSaveClose();
					return TRUE;

				case IDC_BTNQUB : 
					pVwSession->OnBtnQub();
					return TRUE;

				case IDC_BTNCLOSE :
					DestroyWindow( GetParent( hDlg ) );
					return TRUE;	

				//-- Drawing modes 		
				case IDC_BTNDRAWMODE1 :
					pVwSession->SetDrawMode(1);
					return TRUE;	
					
				case IDC_BTNDRAWMODE2 :
					pVwSession->SetDrawMode(2);
					return TRUE;	
					
				case IDC_BTNDRAWMODE3 :
					pVwSession->SetDrawMode(3);
					return TRUE;	
					
				//-- Scan Modes 
				case IDC_BTNSCAN_MODE1 :
					pVwSession->SetScanMode(1);
					return TRUE;	

				case IDC_BTNSCAN_MODE2 :
					pVwSession->SetScanMode(2);
					return TRUE;	
					
				//-- Time separators
				case IDC_BTNTIMESEP : 
					pVwSession->OnBtnTimeSep();
					return true;
					
				//-- click on elapsed time label to reset
				case IDC_LBLTIMEELAPSED : 
					pVwSession->SetElapsed(0);
					return TRUE;
					
				default : 
					if( wmId >= ID_GENCTL+ID_GENCTLX && wmId < ID_GENCTL+ID_GENCTLX*(SESSION_MAXGENCHN+1) ) { 
						InvalidateRect( pVwSession->m_hDlg, &(pVwSession->m_rctDisp), false );
						int iChn = int((wmId-ID_GENCTL-ID_GENCTLX)/ID_GENCTLX);
						int iPnl = pVwSession->m_oSession.m_aiGenChannelPanel[iChn];
						if( pVwSession->m_bSetup 
								&& wmEvent==EN_CHANGE 
								&& iChn>=0 
								&& iPnl<pVwSession->m_oSession.m_aPanels.Count() 
								&& (pVwSession->m_oSession.Panel(iPnl))->m_iChannelType==CT_AO
								&& (pVwSession->m_oSession.Panel(iPnl))->m_pUpdate!=NULL
								&& ! pVwSession->m_bResizing
								) {
							pVwSession->m_oSession.Panel(iPnl)->m_pUpdate->SetButtonBitmap( g_hbUpdate[0] );
							}
						
						}
				}
				
			return FALSE;
					
		case WM_CTLCOLORSTATIC :
			wmId = GetDlgCtrlID( (HWND)lParam );
			if( wmId >= ID_GENCTL && wmId < ID_GENCTL+ID_GENCTLX*(SESSION_MAXGENCHN+1) ) 
				return (BOOL)GetStockObject( WHITE_BRUSH );
			return FALSE;
			
		case WM_PAINT :
			pVwSession->OnPaint();
			return FALSE;
			
		case WM_ACQPOLLSTOP : 
			pVwSession->OnAcqPollerStop();
			return TRUE;
		
		case WM_GENPOLLSTOP : 
			pVwSession->OnGenPollerStop();
			return TRUE;

		case WM_DESTROY : {
			int iPos = WindowListFindID( IDD_ACQNOTE, 0 );
			while( iPos > 0 ) { 
				PostMessage( WindowListHandle( iPos ), WM_ACQNOTE_SESSIONDESTROYED, 0, 0 );
				iPos=WindowListFindID( IDD_ACQNOTE, iPos+1 );
				}
			}
			pVwSession->OnExit();	
			delete pVwSession;
			SetWindowLong( GetParent( hDlg ), 0, 0);
			WindowListRemove( hDlg );
			return TRUE;

		case WM_FINISHED_RESIZING_VWSESS:
			pVwSession->m_bResizing = false;
			break;

		default : 
			break;
		}
		
	return FALSE;
	} 


